(*
  Syntax based type reconstruction
 
  author: Christian Mueller
*)

local
    open AbstractSyntax LCSupport Info
in

exception ReconError of string 
exception UnifyError of mixture list

(* management of free type variables *)

fun int2name x = 
    let 
	fun f x s = 
	    if x <= 0 then (if s="" then "a" else s)
	    else f (x div 26) (s^Char.toString(chr((x mod 26)+ord(#"a"))))
    in
	"'"^ f x ""
    end
		
val (settv,newtv) = 
    let 
	val v = ref ~1 
    in 
	(fn n => v:=n,
	 fn () => (v:=(!v)-1;(emptyInfo,int2name (~(!v)-2),!v))) 
    end

fun resettv () = settv ~1

fun shift_plist _ _    nil        = nil
  | shift_plist c d ((e1,e2)::er) = 
    (shift c d e1, shift c d e2) :: shift_plist c d er


(* subst is different from LambdaCube.subst as it substitutes 
	 free type variables which are independend from deBruijn *)
fun subst (Var (info,s,x))  y e         = if x=y then e else Var (info,s,x)
  | subst (App(info,e1,e2)) y e         = App(info,subst e1 y e, subst e2 y e)
  | subst (Abstr(info,s,e1,e2)) y e     = 
    Abstr(info,s,subst e1 y e, subst e2 y e)

  | subst (Pi(info,s,e1,e2)) y e        = Pi(info,s,subst e1 y e, subst e2 y e)
  (* additional rules for integers and boolean *)
  | subst (IntNeg (info,e1)) y e        = IntNeg(info,subst e1 y e)
  | subst (Plus(info,e1,e2)) y e        = Plus(info,subst e1 y e, subst e2 y e)
  | subst (Minus(info,e1,e2)) y e       = 
    Minus(info,subst e1 y e, subst e2 y e)

  | subst (Mult(info,e1,e2)) y e        = Mult(info,subst e1 y e, subst e2 y e)
  | subst (IntEqual(info,e1,e2)) y e    = 
    IntEqual(info,subst e1 y e, subst e2 y e)

  | subst (IntNotEqual(info,e1,e2)) y e = 
    IntNotEqual(info,subst e1 y e, subst e2 y e)

  | subst (LessEq(info,e1,e2)) y e      = 
    LessEq(info,subst e1 y e, subst e2 y e)

  | subst (Less(info,e1,e2)) y e        = Less(info,subst e1 y e, subst e2 y e)
  | subst (GreaterEq(info,e1,e2)) y e   = 
    GreaterEq(info,subst e1 y e, subst e2 y e)

  | subst (Greater(info,e1,e2)) y e     = 
    Greater(info,subst e1 y e, subst e2 y e)

  | subst (BoolNeg (info,e1)) y e       = BoolNeg(info,subst e1 y e)
  | subst (If(info,e1,e2,e3)) y e       = 
    If(info,subst e1 y e, subst e2 y e, subst e3 y e)       

  (* additional rules for variants *)
  | subst (VariantSort (info,xs)) y e   = 
    VariantSort (info, map (fn (s,e') => (s,subst e' y e)) xs)

  | subst (Variant(info,s,e1,e2)) y e   = 
    Variant (info,s,subst e1 y e, subst e2 y e)

  | subst (Case(info,e1,xs))    y e     = 
    Case(info,subst e1 y e,map (fn (s1,s2,e')=>(s1,s2,subst e' y e)) xs)

  (* additional rules for records *)
  | subst (RecordSort(info,xs)) y e     = 
    RecordSort (info, map (fn (s,e') => (s,subst e' y e)) xs)

  | subst (Record(info,xs)) y e         = 
    Record (info, map (fn (s,e') => (s,subst e' y e)) xs)

  | subst (Proj(info,e',s)) y e         = Proj(info, subst e' y e,s)
  (* additional rules for iso-recursive types *)
  | subst (Fold(info,T,e')) y e         = Fold(info,subst T y e, subst e' y e)
  | subst (Unfold(info,T,e')) y e       = 
    Unfold(info,subst T y e, subst e' y e)

  | subst (RecursiveSort(info,s,T)) y e = RecursiveSort(info,s, subst T y e)
  (* Ascription *)
  | subst (Ascribe(info,e1,e2)) y e     = 
    Ascribe(info,subst e1 y e, subst e2 y e)

  | subst const       _ _               = const


(* substitutes e for y inside a list of constraints *)
fun subst_plist nil           _ _ = nil
  | subst_plist ((e1,e2)::er) y e = 
    (subst e1 y e, subst e2 y e) :: subst_plist er y e

(* applies a substitution to an expression *)
fun apply_subst   nil            e = e
  | apply_subst ((Var(_,_,x),e')::xr) e = apply_subst xr (subst e x e')
  | apply_subst   _   _ = raise ReconError "wrong build substitution"

fun occurs x (Var (_,_,y))            = x=y
  | occurs x (Abstr(_,_,e1,e2)) = occurs x e1 orelse occurs x e2
  | occurs x (Pi(_,_,e1,e2))    = occurs x e1 orelse occurs x e2
  | occurs _  _                 = false

val sort = Listsort.sort (fn ((s1,_),(s2,_)) => String.compare (s1,s2))

(* unifies a list of constraints *)
fun unif done nil = done
  | unif done ((Star _,Star _)::xr)         = unif done xr   
  | unif done ((BoolType _,BoolType _)::xr) = unif done xr
  | unif done ((IntType _,IntType _)::xr)   = unif done xr
  | unif done ((UnitType _,UnitType _)::xr) = unif done xr
  | unif done ((Var(i,s,x), Var (i',s',y))::xr) = 
    if x=y then unif done xr
    else unif ((Var(i,s,x),Var (i',s',y))::(subst_plist done x (Var(i',s',y))))
	(subst_plist xr x (Var (i',s',y)))

  | unif done ((Var(i,s,x),e2)::xr) = 
	if occurs x e2 then 
	    raise UnifyError ([Str "unify error: cyclic constraint ",
			       E (Var(i,s,x)), Str "=", E e2])
	else unif ((Var(i,s,x),e2)::(subst_plist done x e2)) 
	          (subst_plist xr x e2)

  | unif done ((e1,Var v)::xr) = unif done ((Var v,e1)::xr)
  | unif done ((Pi(_,_,e11,e12),Pi(_,_,e21,e22))::xr) = 
		  unif done ((e11,e21)::(e12,e22)::xr)

  | unif done ((RecordSort(_,xs),RecordSort(_,ys))::xr) = 
		  unif done (unif_lists (sort xs) (sort ys) xr)

  | unif _    ((e1,e2)::_)    = raise UnifyError ([Str "cannot not unify ", 
						   E e1, Str " and ",E e2])

and unif_lists nil nil acc = acc 
  | unif_lists ((l1,e1)::xr) ((l2,e2)::yr) acc =  
    if l1=l2 then unif_lists xr yr ((e1,e2)::acc) 
    else raise UnifyError ([Str ("unify: labels \""^l1^"\" and \""^l2^"\""),
			    Str " do not match"])

  | unif_lists _ _ _ = raise UnifyError ([Str "unify: different number ",
					  Str "of components in type list"]) 

fun unify xs = unif nil xs


(* context management of types that are bound to names *)
fun get g name x = 
    shift 0 (x+1) (List.nth(g,x) 
		   handle Subscript => 
		       raise ReconError ("unbound type identifier "^name))

(* 
 constraint typing algorithm
 this algorithm generates a list of constraints which is solved by unification
*)
fun recon g (Star _)     = (Box emptyInfo,nil)
  | recon g (Box _)      = raise ReconError "box found"
  | recon g (IntType _)  = (Star emptyInfo,nil)
  | recon g (BoolType _) = (Star emptyInfo,nil)
  | recon g (UnitType _) = (Star emptyInfo,nil)
  | recon g (Pi (_,_,e1,e2)) = 
    let 
	val (K1,C1) = recon g e1
	val (K2,C2) = recon (K1::g) e2
    in 
	(Star emptyInfo, 
	 List.concat [ [(K1,Star emptyInfo),(K2,Star emptyInfo)], C1, C2 ])
    end

  | recon g (RecordSort (_,xs)) = 
    let 
	fun constrsyn ((l,e),C) = 
	    let 
		val (K,C') = recon g e 
	    in 
		(K,Star emptyInfo)::C'@C 
	    end
    in 
	(Star emptyInfo, foldl constrsyn nil xs)
    end
  | recon g (Int (_,i))  = (IntType emptyInfo,nil)
  | recon g (True _)     = (BoolType emptyInfo,nil)
  | recon g (False _)    = (BoolType emptyInfo,nil)
  | recon g (Unit _)     = (UnitType emptyInfo,nil)
  | recon g (Var(_,name,x))  = (get g name x handle _ => Star emptyInfo, nil)
  | recon g (If(_,e1,e2,e3)) = 
    let
	val (T1,C1) = recon g e1
	val (T2,C2) = recon g e2
	val (T3,C3) = recon g e3
    in
	(T2,List.concat [ [(T2,T3),(T1,BoolType emptyInfo)], C1, C2, C3 ] )
    end
  | recon g (Plus(_,e1,e2))        = 
    reconbinop g e1 e2 (IntType emptyInfo) (IntType emptyInfo)

  | recon g (Minus(_,e1,e2))       = 
    reconbinop g e1 e2 (IntType emptyInfo) (IntType emptyInfo)

  | recon g (Mult(_,e1,e2))        = 
    reconbinop g e1 e2 (IntType emptyInfo) (IntType emptyInfo)

  | recon g (IntEqual(_,e1,e2))    = 
    reconbinop g e1 e2 (IntType emptyInfo) (BoolType emptyInfo)

  | recon g (IntNotEqual(_,e1,e2)) = 
    reconbinop g e1 e2 (IntType emptyInfo) (BoolType emptyInfo)

  | recon g (LessEq(_,e1,e2))      = 
    reconbinop g e1 e2 (IntType emptyInfo) (BoolType emptyInfo)

  | recon g (GreaterEq(_,e1,e2))   = 
    reconbinop g e1 e2 (IntType emptyInfo) (BoolType emptyInfo)

  | recon g (App(_,e1,e2)) = 
    let
	val (T1,C1) = recon g e1
	val (T2,C2) = recon g e2
	val v = newtv ()
    in
	(Var v, List.concat [ [(T1,Pi(emptyInfo,"",T2,Var v))], C1, C2 ])
    end

  | recon g (Abstr(_,_,Box _,e2)) = 
    let
	val v = newtv ()
	val (T,C) = recon (Var v :: g) e2
    in
	(Pi(emptyInfo,"",Var v, T), C)
    end 

  | recon g (Abstr(_,_,e1,e2)) = 
    let
	val (_,C1) = recon g e1 
	val (T,C2) = recon (e1 :: g) e2
	val ty = Pi(emptyInfo,"",e1,T)
	val (_,C3) = recon g ty
    in
	(ty, List.concat ([ C1, C2, C3 ]))
    end 

  | recon g (BoolNeg (_,e)) = 
    let 
	val (T,C) = recon g e 
    in 
	((BoolType emptyInfo),(T,(BoolType emptyInfo))::C) 
    end

  | recon g (IntNeg (_,e))  = 
    let 
	val (T,C) = recon g e 
    in 
	((IntType emptyInfo),(T,(IntType emptyInfo))::C) 
    end

  | recon g (Record(_,xs)) =  
    let
	fun f ((l,e),(xs,cs)) = let val (T,C) = recon g e 
				in ((l,T)::xs,C @ cs)
				end
	val (xs,C) = foldr f (nil,nil) xs
	val ty = RecordSort (emptyInfo,xs)
	val (_,C') = recon g (apply_subst (unify C) ty)
    in
	(ty,C @ C')
    end

  (* this cases differ from the former ones in an important way:
     if the following constructs are wronly typed, recon fails ! *)
  | recon g (Proj(_,e,s))      = 
    (case recon g e of
	 (RecordSort (_,xs),C) => 
	     (case List.find (fn (l,_)=> s=l) xs of
		  NONE => raise ReconError ("no record field \""^s^"\" found")
		| SOME (_,T) => (T,C))
       | _ =>  raise ReconError "wrong build projection" )

  | recon _  _ = raise ReconError "this construct is wrong or not supported"

and reconbinop g e1 e2 argtype resulttype = 
    let
	val (T1,C1) = recon g e1
	val (T2,C2) = recon g e2
    in
	(resulttype, List.concat [ [(T1,argtype),(T2,argtype)], C1, C2 ])
    end
				   
(* 
  This procedure creates a list of constraints and unifies it.
  The resulting substitution is applied to the term.
*)
fun check e =
    let
	val (T,C) = recon nil e
	val s = unify C
    in
	resettv ();
	apply_subst s T
    end						           
end
