(*
 The interpreter module can execute a command list and change the interpreter
 settings for printing and typing. Most of the error handling takes place in
 this file.
 
 author: Christian Mueller
*)

local
    open AbstractSyntax BindingAnalysis LambdaCube Support
    open PrettyPrinter ParserSupport LCSupport Info
in

(* chooses the correct typing function according to the mode *)
fun check  e = case !typingmode of
		   R  => Recon.check e
		 | RG => ReconGraphbased.check e                   
		 | U  => Star emptyInfo
                 | UD => UserCheckEval.check e
		 | _  => LambdaCube.check e

(* choose the correct eval function according to the mode *)
fun eval e = case !typingmode of
                    UD => UserCheckEval.eval e
                  | _  => LambdaCube.eval e 

(* sets the typing mode and modifies LambdaCube settings accordingly *)
fun setMode R    = (typingmode := R;    LambdaCube.setTypingMode R)
  | setMode RG   = (typingmode := RG;   LambdaCube.setTypingMode RG)
  | setMode U    = (typingmode := U;    LambdaCube.setTypingMode U)
  | setMode UD   = (typingmode := UD;   LambdaCube.setTypingMode UD)
  | setMode mode = (typingmode := mode; LambdaCube.setTypingMode mode)

(* choose subtyping or not *)
fun setSub b = subtyping := b

(* clears the environment if necessary for typing mode switch *)
fun adjustEnv env = 
    let
	fun modify ((id,e),(env',c)) =  (check e; (env',c)) 
	    handle _ => (Environment.delete (env',id),true)

	val (env',changed) = foldl modify (env,false) (Environment.getList env)
    in
	if changed then print ", environment modified !!!\n"
	else print "\n";
	env'
    end

(* modifies the settings of the interpreter *)
fun changeSettings env (ModeChange mode) =
    (setPrintMode mode;
     print ("! PP: mode changed to " ^ 
	    (case mode of 
		 AbstractSyntax.Opaque => "opaque\n" 		     
	       | AbstractSyntax.Transparent => "transparent\n"
	       | AbstractSyntax.OpaqueWithSync => 
		     "opaque, with environment sync; may diverge !!!\n"));
     env)		      

  | changeSettings env (OutputChange out) = 
    (setPrintOutput out;
     print ("! PP: output " ^ 
	    (case out of 
		 AbstractSyntax.Term => "terms\n" 
	       | AbstractSyntax.Type => "types\n"
	       | AbstractSyntax.All  => "terms and types\n"));
     env)

  | changeSettings env (TypeModeChange mode) =
    (print ("! mode changed to "^
	    (case mode of S=>"S"|F=>"F"|Fw=>"Fw"|CC=>"CC"
	                 |R=>"R"|RG=>"RG"|U=>"U"|UD=>"UD"));
     setMode mode;
     adjustEnv env)

  | changeSettings env (SubtypingChange b) = 
	if (case !typingmode of S=>true|F=>true|Fw=>true|CC=>true|_=>false) 
	    then (print "! subtyping "; 
		  if b then print "enabled" else print "disabled";
		      setSub b;
		      adjustEnv env)
	else (print "! subtyping not available for this typing mode\n"; env)
	     
(* executes a single command *)
fun executeCommand env (Define (id,e)) =
    let 
	val e' = resolveFreeBinder env e
	val s  = check e'	       
    in  
	pprint ("> def "^id^" = ");
	pplcwredex env e' s;
	Environment.add(env,id,e')
    end	

  | executeCommand env (ReduceDefine (id,e)) =
    let 
	val e'  = resolveFreeBinder env e
	val s   = check e'
	val e'' = eval e'
    in  
	pprint ("> def "^id^" = ");
	if getPrintOutput () = All orelse getPrintOutput () = Term then pp e''
	else pprint "<term>";	
	if getPrintOutput () = All orelse getPrintOutput () = Type 
	    then (pprint " : ";pplc env s) 
	else ();
	ppflush ();
	Environment.add(env,id,e'')
    end

  | executeCommand env (ReduceStep (id,e)) =
    let 
	val e'  = resolveFreeBinder env e	
	val s   = check e'
	val e'' = onestep_eval e' 
	    handle LambdaCube.NoFurtherEval => 
		   (pprint "! no further eval step possible\n"; e')
    in 	
	pprint ("> def "^id^" = "); 
	pplcwredex env e'' s;
	Environment.add(env,id,e'')
    end

  | executeCommand env (Equiv (e1,e2)) = 
    let
	val (e1',e2') = (resolveFreeBinder env e1, resolveFreeBinder env e2)   
	val _ = (check e1', check e2')
    in
	if equiv (e1',e2') then pprint "> equiv: expressions are equivalent" 
	else pprint "> equiv: expressions are NOT equivalent";
	ppflush ();
	env
    end

  | executeCommand env (Protocol e) = 
    let
	val e' = resolveFreeBinder env e	
	val s  = check e'
    in  
	pprint ("> def "^defaultIdent^" = ");
	pplcwredex env e' s;
	showProtocol (Environment.add(env,defaultIdent,e')) e'
    end
  | executeCommand env (Switch s)   = changeSettings env s
  | executeCommand env (UseFile s)  = executeCommandList env (parse_files [s])
				     
and executeCommandList env xs = 
    foldl (fn (c,env') => executeCommand env' c) env xs

and showProtocol env e =
    (let
	 val e' = onestep_eval e 
	 val s  = check e'
     in
	 pprint ("> def "^defaultIdent^" = "); 
	 pplcwredex env e' s;
	 showProtocol (Environment.add(env,defaultIdent,e')) e'
     end) handle LambdaCube.NoFurtherEval => env	

(* print an expression list *)

fun printlst env xs =
    let
	fun pr nil = ()
	  | pr (E e::xr) = (pprint "\""; pplc env e; pprint "\""; pr xr)
	  | pr (Str s::xr) = (pprint s; pr xr)
    in
	pr xs;
	PrettyPrinter.ppflush ()
    end

(* prints a help message on the screen *)

fun help () = 
    print ("The following commands are available:\n\n"
           ^"help|?                          shows this message\n"
	   ^"exit                            exits the interpreter\n"
	   ^"switch tm s|f|fw|cc|u|ud|r|rg   switch typing mode\n"
	   ^"switch pm t|o|ows|terms|types   switch printing mode\n"
	   ^"switch sub|nosub                enables/disables subtyping\n"
	   ^"use \"path_of_a_file\"            loads a file\n"
	   ^"def id = expression             defintion without reduction\n"
	   ^"rdef id = expression            same as def but with reduction\n"
	   ^"\nFinish the command with \";\" and press RETURN to pass it to "
	   ^"the interpreter.\nFor more details have a look at the manual.\n")
	   
(* error handling function for all interpreter errors*)

fun showInfo (line,c,file,_) =
    if line = ~1 orelse c = ~1 then print (file^"\n!")
    else  print (file^": line "
		 ^Int.toString line
		 ^":"
		 ^Int.toString c^"\n! ")

fun errorhandler env exn =
    case exn of
        Fail s => (ppflush (); print ("! error: "^s^"\n\n"))
      | LexerParserError => (ppflush ())
      | Environment.IdNotFound (info,s) => 
	    (ppflush (); print("! error: "); 
	     showInfo info;
	     print ("unkown identifier \""^s^"\"\n\n"))

      | LambdaCube.Error s => (ppflush (); print ("! LC error: "^s^"\n"))
      | Recon.ReconError s => (ppflush (); print ("! Recon error: "^s^"\n"))
      | Recon.UnifyError lst => 
	    (ppflush (); print "! Recon error: "; printlst env lst)

      | ReconGraphbased.UnifyError lst => 
	    (ppflush (); print "! ReconGB error: "; printlst env lst)

      | ReconGraphbased.ReconGBError s => 
	    (ppflush (); print ("! ReconGB error: "^s^"\n"))

      | LambdaCube.ErrorWInfo (info,lst) => 
	    (print "! LC error: "; showInfo info; printlst env lst)

      | Overflow => (ppflush (); 
		     print ("! error: overflow during integer operation\n\n")) 

      | UserCheckEval.UserError s => 
	    (ppflush (); print ("! UD error: "^s^"\n")) 

      | UserCheckEval.UserErrorWInfo (info,lst) => 
	    (ppflush (); print ("! UD error: ");showInfo info;printlst env lst)

      | Support.Help => help ()	    
      | _ => raise exn

(* main loop of the interpreter *)

fun run env parser =
   (resetLineno ();
    print ("- ");     
    run (executeCommandList env (Parser BasicIO.std_in)) parser)
   handle ExitToplevel => (ppflush (); print ("\n\n<-- CU\n"))
	| Lexer.Exit => (ppflush (); print ("<-- CU \n"))	    
	| exn => (errorhandler env exn; run env parser)
	
end		



